/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  23.02.2016  mifi  First Version, tested with an Embedded Artists 
*                    LPC1788 Developer's Kit.
**************************************************************************/
#define __TALCPU_COM_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include "tal.h"
#include "uart_17xx40xx.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/*
 * Use UART FIFO
 *
 *    0 = disabled
 *    1 = enabled
 */
#define USE_UART_FIFO   1

/*
 * Fractional Divider setting look-up table
 * Based on the user manual UM10237, 
 * Rev. 04 - 28 August 2009, p440
 */
typedef struct _fdslut 
{
   uint16_t wFR;
   uint8_t  bDivAddVal;
   uint8_t  bMulVal;
} FDSLUT;

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

static FDSLUT const LookUpTable[72] =
{
   {1000,0,1 }, {1067,1,15 }, {1071,1,14 }, {1077,1,13 }, {1083,1,12 }, {1091,1,11 },
   {1100,1,10}, {1111,1,9  }, {1125,1,8  }, {1133,2,15 }, {1143,1,7  }, {1154,2,13 },
   {1167,1,6 }, {1182,2,11 }, {1200,1,5  }, {1214,3,14 }, {1222,2,9  }, {1231,3,13 },
   
   {1250,1,4 }, {1267,4,15 }, {1273,3,11 }, {1286,2,7  }, {1300,3,10 }, {1308,4,13 },
   {1333,1,3 }, {1357,5,14 }, {1364,4,11 }, {1375,3,8  }, {1385,5,13 }, {1400,2,5  },
   {1417,5,12}, {1429,3,7  }, {1444,4,9  }, {1455,5,11 }, {1462,6,13 }, {1467,7,15 },
   
   {1500,1,2 }, {1533,8,15 }, {1538,7,13 }, {1545,6,11 }, {1556,5,9  }, {1571,4,7  },
   {1583,7,12}, {1600,3,5  }, {1615,8,13 }, {1625,5,8  }, {1636,7,11 }, {1643,9,14 },
   {1667,2,3 }, {1692,9,13 }, {1700,7,10 }, {1714,5,7  }, {1727,8,11 }, {1733,11,15},
   
   {1750,3,4 }, {1769,10,13}, {1778,7,9  }, {1786,11,14}, {1800,4,5  }, {1818,9,11 },
   {1833,5,6 }, {1846,11,13}, {1857,6,7  }, {1867,13,15}, {1875,7,8  }, {1889,8,9  },
   {1900,9,10}, {1909,10,11}, {1917,11,12}, {1923,12,13}, {1929,13,14}, {1933,14,15}   
};

static TAL_COM_DCB *DCBArray[TAL_COM_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  Name  : CalculateFractionalDivider                                   */
/*                                                                       */
/*  In    : dBaudrate, pDivisor, pFracDiv                                */
/*  Out   : pDivisor, pFracDiv                                           */
/*  Return: none                                                         */
/*************************************************************************/
static void CalculateFractionalDivider (uint32_t  dBaudrate, 
                                        uint16_t *pDivisor,
                                        uint16_t *pFracDiv)
{
   uint32_t dPCLK;
   uint32_t dDivisor;
   uint32_t dRemainder;
   uint32_t dRemainderMin;
   uint8_t  bIndex;
   uint8_t  bIndexMax;
   uint8_t  bFinalIndex;
   
   /* Get UART clock */
   dPCLK = tal_CPUGetFrequencyAPB();
   
   /*
    * Because the FR value from the look-up table was
    * multiply by 1000, the divisor must be multiply too.
    */
           
   /* x / 16 / baudrate */  
   dDivisor = ((dPCLK >> 4) * 1000) / dBaudrate;
        
   /*
    * Find the index with the smallest remainder.
    */
   dRemainderMin = 0xffffffff;
   bFinalIndex   = 0;
   bIndexMax     = sizeof(LookUpTable) / sizeof(FDSLUT);
   for (bIndex=0; bIndex<bIndexMax; bIndex++)
   {
      dRemainder = dDivisor % LookUpTable[bIndex].wFR;
      if (dRemainder < dRemainderMin)
      {
         bFinalIndex   = bIndex;
         dRemainderMin = dRemainder;
      }
   } 
   
   /* 
    * bFinalIndex is the entry with the smallest remainder now.
    */       
   *pDivisor = (uint16_t)(dDivisor / LookUpTable[bFinalIndex].wFR);
   *pFracDiv = (uint16_t)((LookUpTable[bFinalIndex].bMulVal << 4) | LookUpTable[bFinalIndex].bDivAddVal);
      
} /* CalculateFractionalDivider */

/*************************************************************************/
/*  RS485TxEnable                                                        */
/*                                                                       */
/*  Enable RS485 TX mode.                                                */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void RS485TxEnable (TAL_COM_DCB *pDCB)
{
   if (pDCB->HW.RS485TXEnable != NULL)
   {
      pDCB->HW.RS485TXEnable();
   }
} /* RS485TxEnable */

/*************************************************************************/
/*  RS485TxDisable                                                       */
/*                                                                       */
/*  Disable RS485 TX mode.                                               */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void RS485TxDisable (TAL_COM_DCB *pDCB)
{
   if (pDCB->HW.RS485TXDisable != NULL)
   {
      pDCB->HW.RS485TXDisable();
   }
} /* RS485TxEnable */

/*************************************************************************/
/*  ReadData                                                             */
/*                                                                       */
/*  Read data from the Receive Buffer Register.                          */
/*                                                                       */
/*  In    : pDCB, pUARTx                                                 */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void ReadData (TAL_COM_DCB *pDCB, UART_TypeDef *pUARTx)
{
   TAL_RESULT  Error;
   uint8_t    bData;

   /* Read data */
   bData = (uint8_t)pUARTx->RBR;
   
   /* If we have no overflow... */
   if (TAL_FALSE == pDCB->bRxOverflow)
   {
      /* ... put it into the ring buffer */
      Error = tal_MISCRingAdd(&pDCB->RxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Signal counting semaphore */
         OSSemaSignalFromInt(&pDCB->RxRdySema);
      }
      else
      {
         /* Ups, overflow */
         pDCB->bRxOverflow = TAL_OK;
      }
   }  

} /* ReadData */

/*************************************************************************/
/*  IRQHandler                                                           */
/*                                                                       */
/*  This is the generic IRQ handler.                                     */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void IRQHandler (TAL_COM_DCB *pDCB)
{
   TAL_RESULT     Error;
   UART_TypeDef *pUARTx;   
   uint32_t      dIIR;
  uint32_t       dIER;
  uint32_t       dLSR;
  uint8_t        bIndex;
  uint8_t        bData;

   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   /* Get IRQ status */   
   dIIR = pUARTx->IIR;
   dLSR = pUARTx->LSR;
   dIER = pUARTx->IER;
   
   /* 
    * Check for RX interrupt
    */
    
   /* Check if FIFO is used */ 
   if (pDCB->HW.bFifoTriggerLevel != 0)
   {
      /*
       * FIFO is used
       */
       
      /* Check for a block of data, FIFO trigger level */
      if      (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         /* A block of data, FIFO trigger level,  can be read */
         for (bIndex = 0; bIndex < pDCB->HW.bFifoTriggerLevel; bIndex++)
         {
            ReadData(pDCB, pUARTx);
         }
      }
      /* Check Character Time-out Indicator */
      else if (UART_IIR_INTID_CTI == (dIIR & UART_IIR_INTID_MASK))
      {
         while (dLSR & UART_LSR_RDR)
         {
            ReadData(pDCB, pUARTx);
            dLSR = pUARTx->LSR;
         }
      }
   }
   else
   {
      /*
       * FIFO is NOT used
       */
       
      /* Check if data is available */
      if (UART_IIR_INTID_RDA == (dIIR & UART_IIR_INTID_MASK))
      {
         ReadData(pDCB, pUARTx);
      } /* end "RX interrupt" */
   }      
   
   
   /* 
    * Check for TX interrupt, but only if enabled 
    */
   if ((dIER & UART_IER_THREINT) &&                            /* <= enabled ? */
       (UART_IIR_INTID_THRE == (dIIR & UART_IIR_INTID_MASK)))  /* <= TX interrupt ? */
   {
      /* Read Data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupt */
         pUARTx->IER &= ~UART_IER_THREINT;;
         
         /* Disable RS485 TX mode */
         RS485TxDisable(pDCB);
      }
      else
      {
         /* Check if FIFO is used */ 
         if (pDCB->HW.bFifoTriggerLevel != 0)
         {
            /*
             * FIFO is used
             */
             
            /* Put first data into the FIFO */
            pUARTx->THR = bData;
            
            /* Get next data, sub one because the first data is in the FIFO */   
            bIndex = pDCB->HW.bFifoTriggerLevel - 1;
            while (bIndex != 0)
            {
               /* Get next data */
               Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
               if (Error != TAL_OK)
               {
                  /* No data available anymore in the TX ring buffer */
                  break;
               }
               else
               {
                  /* Put the data into the FIFO */
                  pUARTx->THR = bData;
               }
               bIndex--;
            }
         }
         else
         {
            /*
             * FIFO is NOT used
             */
            pUARTx->THR = bData; 
         }            
      }   
   } /* end "TX interrupt */
   
} /* IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  USARTx_IRQHandler                                                    */
/*                                                                       */
/*  This is the Cortex USARTx IRQ handler.                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void UART0_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_1]);
   TAL_CPU_IRQ_EXIT();
} /* UART0_IRQHandler */

void UART1_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_2]);
   TAL_CPU_IRQ_EXIT();
} /* UART1_IRQHandler */

void UART2_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_3]);
   TAL_CPU_IRQ_EXIT();
} /* UART2_IRQHandler */

void UART3_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_4]);
   TAL_CPU_IRQ_EXIT();
} /* UART3_IRQHandler */

void UART4_IRQHandler (void)
{
   TAL_CPU_IRQ_ENTER();
   IRQHandler(DCBArray[TAL_COM_PORT_5]);
   TAL_CPU_IRQ_EXIT();
} /* UART4_IRQHandler */

/*************************************************************************/
/*  cpu_COMInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the COM port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMInit (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_ERR_COM_PORT_RANGE;
   TAL_COM_HW  *pHW    = &pDCB->HW;
   
   /* Set default values for RS485 */
   pHW->bRS485Mode     = FALSE;
   pHW->RS485TXEnable  = NULL;
   pHW->RS485TXDisable = NULL;
   
   switch (pDCB->ePort)
   {
      case TAL_COM_PORT_1:
      {
         Error = tal_BoardEnableCOM1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_1] = pDCB;
         
            pHW->dBaseAddress = LPC_UART0_BASE;
            pHW->IrqNumber    = UART0_IRQn;
            pHW->bIrqPriority = UART0_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART0);
         }
         break;
      } /* TAL_COM_PORT_1 */
      
      case TAL_COM_PORT_2:
      {
         Error = tal_BoardEnableCOM2();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_2] = pDCB;
         
            pHW->dBaseAddress = LPC_UART1_BASE;
            pHW->IrqNumber    = UART1_IRQn;
            pHW->bIrqPriority = UART1_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART1);
         }
         break;
      } /* TAL_COM_PORT_2 */
           
      case TAL_COM_PORT_3:
      {
         Error = tal_BoardEnableCOM3();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_3] = pDCB;
         
            pHW->dBaseAddress = LPC_UART2_BASE;
            pHW->IrqNumber    = UART2_IRQn;
            pHW->bIrqPriority = UART2_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART2);
         }
         break;
      } /* TAL_COM_PORT_3 */
      
      case TAL_COM_PORT_4:
      {
         Error = tal_BoardEnableCOM4();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_4] = pDCB;
         
            pHW->dBaseAddress = LPC_UART3_BASE;
            pHW->IrqNumber    = UART3_IRQn;
            pHW->bIrqPriority = UART3_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART3);
         }
         break;
      } /* TAL_COM_PORT_4 */
      
      case TAL_COM_PORT_5:
      {
         Error = tal_BoardEnableCOM5();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_COM_PORT_5] = pDCB;
         
            pHW->dBaseAddress = LPC_UART4_BASE;
            pHW->IrqNumber    = UART4_IRQn;
            pHW->bIrqPriority = UART4_PRIO;

            /* Reset UART */
            Chip_SYSCTL_PeriphReset(SYSCTL_RESET_UART4);
         }
         break;
      } /* TAL_COM_PORT_5 */
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_COMInit */

/*************************************************************************/
/*  cpu_COMIoctl                                                         */
/*                                                                       */
/*  Call a IOCTL function.                                               */
/*                                                                       */
/*  In    : pDCB, wNum, pParam                                           */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMIoctl (TAL_COM_DCB *pDCB, TAL_COM_IOCTL eFunc, uint32_t *pParam)
{
   TAL_RESULT Error = TAL_OK;
   
   switch (eFunc)
   {
      case TAL_COM_IOCTL_RS485_MODE:
      {
         pDCB->HW.bRS485Mode = (uint8_t)*pParam;
         break;
      }   
         
      case TAL_COM_IOCTL_RS485_TX_ENABLE_FUNC:
      {
         pDCB->HW.RS485TXEnable = (void(*)(void))*pParam;      
         break;
      }
      
      case TAL_COM_IOCTL_RS485_TX_DISABLE_FUNC:
      {
         pDCB->HW.RS485TXDisable = (void(*)(void))*pParam;      
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_IOCTL_NUM;
         break;
      }
   } /* end switch (wNum) */

   return(Error);
} /* cpu_COMIoctl */

/*************************************************************************/
/*  cpu_COMOpen                                                          */
/*                                                                       */
/*  Open the COM port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMOpen (TAL_COM_DCB *pDCB)
{
   TAL_RESULT      Error = TAL_ERROR;
   UART_TypeDef  *pUARTx;   
   uint32_t       dLCR   = 0;
   uint8_t        bDummy = 0;
   uint16_t       wDivisor;
   uint16_t       wFracDiv;  
   
   /* Disable FIFO first */   
   pDCB->HW.bFifoTriggerLevel = 0;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;
         
   /* Check word length */
   switch (pDCB->Settings.eLength)
   {
      case TAL_COM_LENGTH_8:
      {
         dLCR |= UART_LCR_WLEN8;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_LENGTH;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eLength) */
   
   /* Check parity settings */
   switch (pDCB->Settings.eParity)
   {
      case TAL_COM_PARITY_NONE:
      {
         dLCR |= UART_LCR_PARITY_DIS;  /*lint !e845 */
         break;
      } 
      
      case TAL_COM_PARITY_EVEN:
      {
         dLCR |= UART_LCR_PARITY_EVEN;
         break;
      }
      
      case TAL_COM_PARITY_ODD:
      {
         dLCR |= UART_LCR_PARITY_ODD;  /*lint !e845 */
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_PARITY;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eParity) */
    
   /* Check stop bit settings */
   switch (pDCB->Settings.eStop)
   {
      case TAL_COM_STOP_1_0:
      {
         dLCR |= UART_LCR_SBS_1BIT;    /*lint !e845 */
         break;
      }
      
      case TAL_COM_STOP_2_0:
      {
         dLCR |= UART_LCR_SBS_2BIT;
         break;
      }
      
      default:
      {
         Error = TAL_ERR_COM_STOP;
         goto COMOpenEnd;  /*lint !e801*/
         break;   /*lint !e527*/
      }
   } /* switch (pDCB->Settings.eStop) */

   /* Check baud rate */
   if (pDCB->Settings.dBaudrate != 0)
   {
      CalculateFractionalDivider(pDCB->Settings.dBaudrate, &wDivisor, &wFracDiv);
   }
   else
   {
      Error = TAL_ERR_COM_BAUDRATE;
      goto COMOpenEnd;  /*lint !e801*/
   }

   /* Disable RS485 TX => Enable RX */   
   RS485TxDisable(pDCB);

   /* 
    * Set UART register 
    */
    
   /* Disable all interrupts first */ 
   pUARTx->IER = 0x00;
   
   /* Set baudrate */
   pUARTx->LCR = UART_LCR_DLAB_EN;  /* Enable access to Divisor Latches */  
   pUARTx->DLL = (wDivisor & 0xFF);
   pUARTx->DLM = ((wDivisor >> 8) & 0xFF);
   pUARTx->FDR = wFracDiv;
   pUARTx->LCR = 0;                 /* Disable access to Divisor Latches */
   
   /* Set line control register */
   pUARTx->LCR = dLCR;
   
   /* Set FIFO and Auto-baud Control */
   pUARTx->FCR = 0;
   pUARTx->ACR = 0;

#if (USE_UART_FIFO >= 1)
   /* Reset and enable FIFOs, FIFO trigger level 2 (8 chars) */
   pDCB->HW.bFifoTriggerLevel = 8;                        
   
   pUARTx->FCR = (UART_FCR_FIFO_EN | 
                  UART_FCR_RX_RS   | 
                  UART_FCR_TX_RS   | 
                  UART_FCR_TRG_LEV2);
#endif                        
                        
   /* Dummy read to clear the RBR register */
   while (UART_LSR_RDR & pUARTx->LSR)
   {
      bDummy = (uint8_t)pUARTx->RBR;
   }  
   (void)bDummy;
   
   /* Enable RX interrrupt */
   pUARTx->IER |= UART_IER_RBRINT;
   
   /* Set priority and enable the IRQ */
   tal_CPUIrqSetPriority(pDCB->HW.IrqNumber, pDCB->HW.bIrqPriority);
   tal_CPUIrqEnable(pDCB->HW.IrqNumber);

   Error = TAL_OK;
   
COMOpenEnd:   
   return(Error);
} /* cpu_COMOpen */

/*************************************************************************/
/*  cpu_COMClose                                                         */
/*                                                                       */
/*  Close the COM port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMClose (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   UART_TypeDef *pUARTx;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;
   
   /* Disable the RS485 TX mode */
   RS485TxDisable(pDCB);

   /* Disable the IRQ */   
   tal_CPUIrqDisable(pDCB->HW.IrqNumber);
   
   /* Disable TX and RX interrupt  */
   pUARTx->IER = 0;
   
   return(Error);
} /* cpu_COMClose */

/*************************************************************************/
/*  cpu_COMStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_COMStartTx (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   UART_TypeDef *pUARTx;
   uint8_t      bData;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUARTx->IER & UART_IER_THREINT)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      /* Get data from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, &bData);
      if (TAL_OK == Error)
      {
         /* Enable RS485 TX mode */
         RS485TxEnable(pDCB);
      
         /* Send data */
         pUARTx->THR = bData;
      
         /* Enable TX interrupt */
         pUARTx->IER |= UART_IER_THREINT;
      }               
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMStartTx */

/*************************************************************************/
/*  cpu_COMTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_COMTxIsRunning (TAL_COM_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   UART_TypeDef *pUARTx;
   
   /* Get UART pointer */
   pUARTx = (UART_TypeDef*)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pUARTx->IER & UART_IER_THREINT)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_ERROR;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_COMTxIsRunning */

/*** EOF ***/
